
#include <zephyr/kernel.h>
#include <net/coap_utils.h>
#include <zephyr/logging/log.h>
#include <zephyr/net/openthread.h>
#include <zephyr/net/socket.h>
#include <openthread/thread.h>

#include "coap_client_utils.h"

LOG_MODULE_REGISTER(coap_client_utils, CONFIG_COAP_CLIENT_UTILS_LOG_LEVEL);

#define COAP_PORT 5683

#define TEST_URI_PATH 						      "hello"
//also tried "coap://coap.me:5683/hello"

static bool is_connected;

static struct k_work test_msg_work;
static struct k_work on_connect_work;
static struct k_work on_disconnect_work;


/* Options supported by the server */
static const char *const test_option[] = { TEST_URI_PATH, NULL };

/* Destination address */
static struct sockaddr_in6 dest_addr = {
	.sin6_family = AF_INET6,
	.sin6_port = htons(COAP_PORT),
	.sin6_addr.s6_addr = { 0x00, 0x64, 0xff, 0x9b, 0x00, 0x00, 0x00, 0x00,
			       0x00, 0x00, 0x00, 0x00, 0x86, 0x66, 0xda, 0x12 },
	.sin6_scope_id = 0U
};


static void printIPv6Address(const otIp6Address addr)
{
  LOG_INF("Mesh Local EID Address: %x%x.%x%x.%x%x.%x%x.%x%x.%x%x.%x%x.%x%x",
        addr.mFields.m8[0],
        addr.mFields.m8[1],
        addr.mFields.m8[2],
        addr.mFields.m8[3],
        addr.mFields.m8[4],
        addr.mFields.m8[5],
        addr.mFields.m8[6],
        addr.mFields.m8[7],
        addr.mFields.m8[8],
        addr.mFields.m8[9],
        addr.mFields.m8[10],
        addr.mFields.m8[11],
        addr.mFields.m8[12],
        addr.mFields.m8[13],
        addr.mFields.m8[14],
        addr.mFields.m8[15]);
}

static void printUnicastAddresses(struct openthread_context *ot)
{
  const otNetifAddress *n;

  n = otIp6GetUnicastAddresses(ot->instance);
  while (n != NULL)
  {
    printIPv6Address((const otIp6Address)n->mAddress);
    n = n->mNext;
  };
}

static int msg_reply(const struct coap_packet *response, struct coap_reply *reply, const struct sockaddr *from)
{
	ARG_UNUSED(reply);
	ARG_UNUSED(from);
	LOG_INF("GOT %d bytes -> %s", response->max_len, response->data);
	
	return 1;
}

static void send_test_msg(struct k_work *item)
{
	int ret = 0;

	ARG_UNUSED(item);

  ret = coap_send_request(COAP_METHOD_GET,
      (const struct sockaddr *)&dest_addr,
      test_option, NULL, 0, msg_reply);
  LOG_INF("Send test message. ret:%d", ret);
}

static void on_thread_state_changed(uint32_t flags, void *context)
{
	struct openthread_context *ot_context = context;
	
	LOG_INF("Thread state changed: %x", flags);

	if (flags & OT_CHANGED_THREAD_ROLE) {
		switch (otThreadGetDeviceRole(ot_context->instance)) {
		case OT_DEVICE_ROLE_CHILD:
		case OT_DEVICE_ROLE_ROUTER:
		case OT_DEVICE_ROLE_LEADER:
		{
			k_work_submit(&on_connect_work);
			is_connected = true;
			printUnicastAddresses(ot_context);
			break;
		}

		case OT_DEVICE_ROLE_DISABLED:
		case OT_DEVICE_ROLE_DETACHED:
		default:
			k_work_submit(&on_disconnect_work);
			is_connected = false;
			break;
		}
	} 
	else if (flags & OT_CHANGED_JOINER_STATE) {
		LOG_INF("Joiner state changed");
	}
}

static void submit_work_if_connected(struct k_work *work)
{
	if (is_connected) {
		k_work_submit(work);
	} else {
		LOG_INF("Connection is broken");
	}
}

void coap_client_utils_init(ot_connection_cb_t on_connect,
			    ot_disconnection_cb_t on_disconnect)
{
	LOG_INF("INIT ALL!!!");
	
	coap_init(AF_INET6, NULL);

	k_work_init(&on_connect_work, on_connect);
	k_work_init(&on_disconnect_work, on_disconnect);
	k_work_init(&test_msg_work, send_test_msg);

	openthread_set_state_changed_cb(on_thread_state_changed);
}

void coap_client_join_network(void)
{
	openthread_start(openthread_get_default_context());
}

void coap_client_send_test_msg(void)
{
	submit_work_if_connected(&test_msg_work);
}
